#!/usr/bin/env python3

import pytest
from tidyexc.views import *

def test_data_view():
    d = [{'a': 1, 'b': 2}, {'b': 3, 'c': 4}]
    v = data_view(d)

    assert repr(v) == "data_view([{'a': 1, 'b': 2}, {'b': 3, 'c': 4}])"
    assert v == {'a': 1, 'b': 3, 'c': 4}
    assert len(v) == 3
    assert set(v) == {'a', 'b', 'c'}

    assert 'a' in v
    assert 'd' not in v

def test_data_view_get_set_del():
    d = [{'a': 1}, {'b': 2}]
    v = data_view(d)

    assert v.a == v['a'] == 1
    assert v.b == v['b'] == 2

    with pytest.raises(AttributeError):
        v.z
    with pytest.raises(KeyError):
        v['z']

    # attribute syntax
    v.a = 3
    v.b = 4

    assert v.a == v['a'] == 3
    assert v.b == v['b'] == 4
    assert d == [{'a': 1}, {'a': 3, 'b': 4}]

    del v.a
    del v.b

    with pytest.raises(AttributeError):
        del v.z

    assert v.a == v['a'] == 1
    assert d == [{'a': 1}, {}]

    with pytest.raises(AttributeError):
        del v.a
    with pytest.raises(AttributeError):
        del v.b
    
    # item syntax
    v['a'] = 3
    v['b'] = 4

    assert v.a == v['a'] == 3
    assert v.b == v['b'] == 4
    assert d == [{'a': 1}, {'a': 3, 'b': 4}]

    del v['a']
    del v['b']

    with pytest.raises(KeyError):
        del v['z']

    assert v.a == v['a'] == 1
    assert d == [{'a': 1}, {}]

    with pytest.raises(KeyError):
        del v['a']
    with pytest.raises(KeyError):
        del v['b']

def test_nested_data_view():
    d = [{'a': 1, 'b': 2}, {'b': 3, 'c': 4}]
    v = nested_data_view(d)

    assert repr(v) == "nested_data_view([{'a': 1, 'b': 2}, {'b': 3, 'c': 4}])"

    assert v['a'] == [1]
    assert v['b'] == [2, 3]
    assert v['c'] == [4]

    with pytest.raises(KeyError):
        v['z']

    assert v.a == [1]
    assert v.b == [2, 3]
    assert v.c == [4]

    with pytest.raises(AttributeError):
        v.z

    assert v['a','b'] == [(1, 2), (1, 3)]
    assert v['b','a'] == [(2, 1), (3, 1)]
    assert v['b','c'] == [(3, 4)]
    assert v['c','b'] == [(4, 3)]
    assert v['a','c'] == [(1, 4)]
    assert v['c','a'] == [(4, 1)]

    assert v['a','b','c'] == [(1, 3, 4)]

    with pytest.raises(KeyError):
        v['a','z']

    assert v.flatten(0) == {'a': 1, 'b': 2}
    assert v.flatten(1) == {'a': 1, 'b': 3, 'c': 4}
    assert v.flatten(-1) == {'a': 1, 'b': 3, 'c': 4}

def test_info_view():
    p = [(0, 'a'), (1, 'b')]
    v = info_view(p)

    assert repr(v) == "info_view([(0, 'a'), (1, 'b')])"
    assert v[0] == 'a'
    assert len(v) == 2
    assert v.layers() == [(0, 'a'), (1, 'b')]
    assert v == ['a', 'b']

    v[1] = 'c'
    assert v[1] == 'c'
    assert p == [(0, 'a'), (-1, 'c')]

    del v[0]
    assert p == [(-1, 'c')]

    v.append('d')
    assert p == [(-1, 'c'), (-1, 'd')]

    v[:] = ['e', 'f']
    assert p == [(-1, 'e'), (-1, 'f')]

def test_reverse_view():
    l = [1, 2]
    v = reverse_view(l)

    assert repr(v) == '[2, 1]'
    assert len(v) == 2
    assert v[0] == 2
    assert v[1] == 1
    assert list(v) == [2, 1]




from tidyexc import Error

class ParseError(Error):
    pass

def parse_xy_coords(path):
    with ParseError.add_info("path: {path}", path=path):
        lines = _lines_from_file(path)
        coords = []

        for i, line in enumerate(lines, 1):
            with ParseError.add_info("line #{i}: {line}", i=i, line=line):
                coord = _coord_from_line(line)
                coords.append(coord)

        return coords

def _lines_from_file(path):
    try:
        with open(path) as f:
            return f.readlines()

    except FileNotFoundError:
        err = ParseError("can't read file")
        err.hints += "Double-check that the given path actually exists."
        raise err from None

def _coord_from_line(line):
    fields = line.split()

    if len(fields) != 2:
        raise ParseError(
                lambda e: f"expected 2 fields, found {len(e.fields)}",
                fields=fields,
        )

    coord = []

    for field in fields:
        try:
            coord.append(float(field))

        except ValueError:
            raise ParseError("expected a number, not {field!r}", field=field) from None

    return tuple(coord)

##################################### >8 #####################################

import pytest
from pathlib import Path

def test_parse_xy_coords(tmp_path):
    p = tmp_path / 'input.xy'
    p.write_text("1 2\n3 4")
    assert parse_xy_coords(p) == [(1, 2), (3, 4)]

def test_parse_xy_coords_err_1(tmp_path):
    p = tmp_path / 'input.xy'

    with pytest.raises(ParseError) as err:
        parse_xy_coords(p)

    assert err.match(r"path: .*input\.xy")
    assert err.match(r"can't read file")

def test_parse_xy_coords_err_2(tmp_path):
    p = tmp_path / 'input.xy'
    p.write_text("1")

    with pytest.raises(ParseError) as err:
        parse_xy_coords(p)

    assert err.match(r"path: .*input\.xy")
    assert err.match(r"line #1: 1")
    assert err.match(r"expected 2 fields, found 1")

def test_parse_xy_coords_err_3(tmp_path):
    p = tmp_path / 'input.xy'
    p.write_text("1 b")

    with pytest.raises(ParseError) as err:
        parse_xy_coords(p)

    assert err.match(r"path: .*input\.xy")
    assert err.match(r"line #1: 1")
    assert err.match(r"expected a number, not 'b'")



#!/usr/bin/env python3

import textwrap
from shutil import get_terminal_size
from contextlib import contextmanager
from collections import ChainMap
from traceback import format_exc
from .views import data_view, nested_data_view, info_view
from .utils import list_iadd, property_iadd, eval_template

_info_stack = []

class Error(Exception):
    """
    A base class for making exceptions that:

    - Provide direct access to any parameters that are relevant to the error.
    - Use bullet-points to clearly and comprehensively describe the error.

    Example:

    .. code-block:: pycon

        >>> from tidyexc import Error
        >>> class CheeseShopError(Error):
        ...     pass
        ...
        >>> err = CheeseShopError(
        ...         product_name="Red Leicester",
        ...         num_requested=1,
        ...         num_available=0,
        ... )
        >>> err.brief = "insufficient inventory to process request"
        >>> err.info += "{num_requested} {product_name} requested"
        >>> err.blame += "{num_available} available"
        >>> raise err
        Traceback (most recent call last):
          ...
        tidyexc.exc.CheeseShopError: insufficient inventory to process request
        • 1 Red Leicester requested
        ✖ 0 available
    """

    info_bullet = '• '
    "The prefix to use for each `info` message in the formatted error."
    blame_bullet = '✖ '
    "The prefix to use for each `blame` message in the formatted error."
    hint_bullet = '• '
    "The prefix to use for each `hints` message in the formatted error."

    @classmethod
    @contextmanager
    def add_info(cls, *messages, **kwargs):
        """
        Add the given `info` to any exceptions derived from this class that are 
        instantiated inside a with-block.

        A simple example:

        .. code-block:: pycon

            >>> from tidyexc import Error
            >>> with Error.add_info('a: {a}', a=1):
            ...     raise Error()
            ...
            Traceback (most recent call last):
              ...
            tidyexc.exc.Error:
            • a: 1

        The purpose of this function is to make it easier to add contextual 
        information to exceptions.  This is easiest to illustrate with an 
        example.  The following function parses a list of (x, y) coordinates 
        from a file.  Each coordinate appears on its own line and must consist 
        of exactly two whitespace-separated numbers.  There are several 
        different errors this function should detect, but all of the error 
        messages should include the file path, and most should also include the 
        offending line number:

        .. literalinclude:: /parse_xy_coords.py
            :language: python
            :end-before: ##################################### >8 #####################################

        Using `add_info()` simplifies the above code in two major ways:

        - Each piece of contextual information is specified just once and used 
          by multiple exceptions.  There's no way to accidentally raise an 
          exception without this information.

        - The helper functions that actually raise the exceptions don't need to 
          have access to any of the contextual information.  Without 
          `add_info()`, it would either be necessary to pass extra arguments 
          around or to catch and re-raise the exceptions.

        It is possible to nest any number of these context managers.  The 
        `info` bullet points will appear in the order the context managers were 
        invoked.  It is considered good practice for the message templates to 
        only make use of the *kwargs* parameters provided to the same context 
        manager, but templates can access all previously defined parameters.  
        They cannot access any subsequently defined parameters.  If a parameter 
        is defined multiple times, the most recent previous value will be used. 
        For example:

        .. code-block:: pycon

            >>> # This template cannot use the `c` parameter, because it has
            >>> # not been defined yet.  Note that the `b` parameter keeps its 
            >>> # value, even though it is subsequently redefined.
            >>> with Error.add_info('a={a} b={b}', a=1, b=2):
            ...
            ...     # The `a` parameter can be used in this template because 
            ...     # it was defined previously.  The `b` parameter shadows the
            ...     # previous value.
            ...     raise Error('a={a} b={b} c={c}', b=3, c=4)
            ...
            Traceback (most recent call last):
              ...
            tidyexc.exc.Error: a=1 b=3 c=4
            • a=1 b=2

        Because templates cannot be affected by subsequent parameters, it is 
        safe to use `add_info()` in recursive functions, where the same exact 
        parameters might be specified multiple times with different values.  
        The `data` attribute provides access to the current value of each 
        parameter.  The `nested_data` attribute, in contrast, provides access 
        to all values for each parameter.
        """
        try:
            cls.push_info(*messages, **kwargs)
            yield
        finally:
            cls.pop_info()

    @classmethod
    def push_info(cls, *messages, **kwargs):
        """
        Add the given `info` to any exceptions derived from this class that are 
        subsequently instantiated.

        See `add_info()` for more information.  This function is similar, 
        except that it is not a context manager.  That means that you must 
        manually call `pop_info()` or `clear_info()` after calling this 
        function.

        Example:

        .. code-block:: pycon

            >>> from tidyexc import Error
            >>> Error.push_info('a: {a}', a=1)
            >>> try:
            ...     raise Error()
            ... finally:
            ...     Error.pop_info()
            ...
            Traceback (most recent call last):
              ...
            tidyexc.exc.Error:
            • a: 1
        """

        _info_stack.append((cls, messages, kwargs))

    @classmethod
    def pop_info(cls):
        """
        Stop adding the `info` that was most-recently "pushed" to exceptions 
        derived from this class.

        This is the opposite of `push_info()`.
        """
        for i, (cls_i, _, _) in reversed(list(enumerate(_info_stack))):
            if cls_i is cls:
                break
        else:
            raise IndexError(f"no info to pop for {cls}")

        del _info_stack[i]

    @classmethod
    def clear_info(cls):
        """
        Stop adding any `info` that was "pushed" to exceptions derived from 
        this class.
        """
        global _info_stack
        _info_stack = [x for x in _info_stack if x[0] is not cls]

    def put_info(self, *messages, **kwargs):
        i = len(self._data)
        self._info += [(i, m) for m in messages]
        self._data.append(kwargs)


    def __init__(self, brief="", **kwargs):
        """
        Create a new exception.

        You can specify a `brief` message template and any number of parameters 
        (via keyword arguments) when constructing an `Error` instance.  This 
        may be enough for simple exceptions, but most of the time you would 
        subsequently add `info` and `blame` message templates (which cannot be 
        directly specified in the constructor).
        
        Any class-wide parameters and/or message templates specified using 
        `add_info()` or `push_info()` are read and applied during the 
        constructor.  This means that these class-wide functions have no effect 
        on exceptions that have already been instantiated.

        Example:

        .. code-block:: pycon

            >>> raise Error("a: {a}", a=1)
            Traceback (most recent call last):
              ...
            tidyexc.exc.Error: a: 1

        """

        # `info_stack` is in the order that the messages will appear in, i.e.  
        # oldest first.  This is a little inconvenient for `data_view`, because 
        # it means that `ChainMap` naturally reads in the wrong direction, but 
        # `data_view` handles this internally using `reverse_view`.

        self._brief = brief
        self._info = []
        self._blame = list_iadd()
        self._hints = list_iadd()
        self._data = []

        for cls, msgs, kws in _info_stack:
            if isinstance(self, cls):
                self.put_info(*msgs, **kws)

        self._data.append(kwargs)
        self._ctor_data = kwargs

        self._info_view = info_view(self._info)
        self._data_view = data_view(self._data)
        self._nested_data_view = nested_data_view(self._data)

    def __str__(self):
        """
        Return the formatted error message.

        All parameter substitutions are performed at this point, so any changes 
        made to either the message templates or the data themselves will be 
        reflected in the resulting error message.
        """

        # By default, exceptions that occur while printing an exception are 
        # totally sequestered; you just get a message saying "<exception str() 
        # failed>".  This isn't very helpful, so here we manually format the 
        # stack trace and display it to the user.
        try:
            message = ""
            parts = [
                    ('',  [self.brief_str]),
                    (self.info_bullet, self.info_strs),
                    (self.blame_bullet, self.blame_strs),
                    (self.hint_bullet, self.hint_strs),
            ]

            for bullet, strs in parts:
                b = len(bullet)

                for s in strs:
                    s = textwrap.indent(s, b*' ')
                    s = bullet + s[b:]
                    message += s + '\n'

            return message

        except Exception as err:
            # Using format_exc() seems to sometimes trigger stack overflows 
            # within the python interpreter.
            return f"\n\nError occurred while formatting {self.__class__.__name__}:\n\n{err.__class__.__name__}: {err}\n\n"

    @property
    def brief(self):
        """
        A message template briefly describing the error.

        The `tidyverse style guide`__ recommends using the verb "must" when the 
        cause of the problem is pretty clear, and the verb "can't" when it's 
        not.  It's common for this template to be a fixed string (i.e. no 
        parameter substitution), and for the `info` and `blame` templates to 
        reference the parameters of the exception.

        __ https://style.tidyverse.org/error-messages.html

        See `info` for a description of message templates in general.  Unlike 
        `info`, `blame`, and `hints`, there can only be a single `brief` 
        message template.  For this reason, use the assignment operator the set 
        this template (as opposed to the in-place addition operator).

        Example:

        .. code-block:: pycon

            >>> err = Error(a=1)
            >>> err.brief = "a: {a}"
            >>> raise err
            Traceback (most recent call last):
              ...
            tidyexc.exc.Error: a: 1

        """
        return self._brief

    @brief.setter
    def brief(self, value):
        self._brief = value
    
    @property
    def brief_str(self):
        """
        The `brief` message, with all parameter substitution performed.
        """
        return eval_template(self.brief, self.data)

    @property_iadd
    def info(self):
        """
        Message templates describing the context in which the error occurred.

        For example, imagine an error that was triggered because an unmatched 
        brace was encountered when parsing a file.  A good `info` message for 
        this exception might specify the file name and line number where the 
        error occurred.

        A message template can either be a string or a callable:

        - *str*: When the error message is generated, the `str.format` method 
          will be called on the string as follows: ``s.format(**self.data)``.  
          This means that any parameters associated with the exception can be 
          substituted into the message.

        - *callable*: When the error message is generated, the callable will be 
          invoked with `data` as the only argument.  It should return a string, 
          which will be taken as the message.  A common use-case for callable 
          template is to specify f-strings via lambda functions.  This is a 
          succinct way to format parameters using arbitrary expressions (see 
          example below).
          
        The `info`, `blame`, and `hints` attributes are all list-like objects 
        containing message templates.  Special syntax is added such that you 
        can use the ``+=`` operator to add message templates to any of these 
        lists.  You can also use any of the usual list methods to modify the 
        list in-place, although you cannot overwrite these attributes 
        altogether.

        The `info` attribute alone has an additional method called ``layers()`` 
        that returns each info template paired with the index that can be 
        passed to ``nested_data.flatten()`` to get the parameters associated 
        with that particular template.
        
        Example:

        .. code-block:: pycon

            >>> err = Error(a=1, b=[2,3])
            >>> err.info += "a: {a}"
            >>> err.info += lambda e: f"b: {','.join(map(str, e.b))}"
            >>> raise err
            Traceback (most recent call last):
              ...
            tidyexc.exc.Error:
            • a: 1
            • b: 2,3

        """
        return self._info_view

    @property
    def info_strs(self):
        """
        The `info` messages, with all parameter substitution performed.
        """
        return [
                eval_template(x, self.nested_data.flatten(i))
                for i, x in self._info
        ]

    @property_iadd
    def blame(self):
        """
        Message templates describing the specific cause of the error.

        For example, imagine an error that was triggered because an unmatched 
        brace was encountered when parsing a file.  A good `blame` message for 
        this exception would clearly state that an unmatched brace was the 
        cause of the problem.

        See `info` for a description of message templates in general.

        Example:

        .. code-block:: pycon

            >>> from tidyexc import Error
            >>> err = Error(a=1)
            >>> err.blame += "a: {a}"
            >>> raise err
            Traceback (most recent call last):
              ...
            tidyexc.exc.Error:
            ✖ a: 1

        """
        return self._blame

    @property
    def blame_strs(self):
        """
        The `blame` messages, with all parameter substitution performed.
        """
        return [eval_template(x, self.data) for x in self.blame]

    @property_iadd
    def hints(self):
        """
        Message templates suggesting how to fix the error.

        For example, imagine an error that was triggered because some user 
        input didn't match any of the expected keywords.  A good hint for this 
        exception might suggest the expected keyword that was most similar to 
        what the user inputted.

        See `info` for a description of message templates in general.

        Example:

        .. code-block:: pycon

            >>> from tidyexc import Error
            >>> err = Error(a=1)
            >>> err.hints += "a: {a}"
            >>> raise err
            Traceback (most recent call last):
              ...
            tidyexc.exc.Error:
            • a: 1

        """
        return self._hints

    @property
    def hint_strs(self):
        """
        The `hints` messages, with all parameter substitution performed.
        """
        return [eval_template(x, self.data) for x in self.hints]

    @property
    def data(self):
        """
        Parameters relevant to the error.

        This attribute is a dictionary-like object that allows parameters to be 
        accessed either as attributes or as dictionary elements:

        .. code-block:: pycon

            >>> e = Error(a=1)
            >>> e.data.a
            1
            >>> e.data['a']
            1

        If a parameter has been defined multiple times (e.g. with 
        `add_info()`), the most recent value is the one that will be used:

        .. code-block:: pycon

            >>> with Error.add_info(a=1):
            ...     e = Error(a=2)
            ...     e.data.a
            2

        """
        return self._data_view

    @data.setter
    def data(self, values):
        self._ctor_data.clear()
        self._ctor_data.update(values)

    @property
    def nested_data(self):
        """
        A view providing access to *all* values defined for each parameter.

        This is useful when trying to extract information from an exception 
        where some parameters may have been defined multiple times, e.g. if 
        `add_info()` was used in a recursive function.  

        The simplest way to use this view is to access a parameter name either 
        as an attribute or a dictionary element.  This will return a list of 
        all the values associated with that parameter:

        .. code-block:: pycon

            >>> with Error.add_info(a=1):
            ...     e = Error(a=2)
            ...     e.nested_data.a
            [1, 2]

        The ``[]`` operator will also accept a tuple of parameter names, in 
        which case it will return a list of those parameters in every context 
        in which at least one of those parameters was defined.  This is useful 
        if you're interested in parameters that are logically connected (e.g. 
        line and column number) and you want to avoid the possibility of them 
        getting out of sync:

        .. code-block:: pycon

            >>> with Error.add_info(a=1, b=2):
            ...     with Error.add_info(a=3, b=4):
            ...         e = Error(c=5)
            ...         e.nested_data['a','b']
            [(1, 2), (3, 4)]

        Finally, the view also has a ``flatten()`` method that can be used to 
        get all the values for each parameter defined at a particular point in 
        time.  The method accepts a single argument which will be used as an 
        index into the internal list of contexts.  The ``data.layers()`` method 
        can be used to get the index corresponding to any info message 
        template.
        """
        return self._nested_data_view





#!/usr/bin/env python3

import functools
from traceback import format_exc

class only_raise:
    """
    Guarantee that the decorated function can only raise the given type of 
    exception.

    Any unhandled exception raised by the decorated function will be caught and 
    re-raised using an exception of the given type. 
    """

    def __init__(self, err_cls):
        self.err_cls = err_cls

    def __call__(self, f):

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except self.err_cls as err:
                raise err from None
            except Exception as err:
                raise self.err_cls(str(err)) from err

        return wrapper


class iadd_mixin:

    def __iadd__(self, other):
        if callable(other) or isinstance(other, str):
            self.append(other)
        else:
            self.extend(other)
        return self

class list_iadd(iadd_mixin, list):
    pass

def property_iadd(getter):
    # Provide a more helpful error message if the user forgets to use `+=`.

    def setter(self, x):
        if x is not getter(self):
            raise AttributeError("can't set attribute, did you mean to use `+=`?")

    return property(getter, setter)

def eval_template(template, data):
    if callable(template):
        return template(data)
    else:
        return template.format(**data)



#!/usr/bin/env python3

from collections import ChainMap
from collections.abc import Sequence, MutableSequence, MutableMapping
from .utils import iadd_mixin

class data_view(MutableMapping):

    def __init__(self, data_stack):
        # Don't trigger `__setattr__()`, and use name-mangling to avoid 
        # clashing with user-set attributes.
        self.__dict__['_data_view__data'] = ChainMap()
        self.__data.maps = reverse_view(data_stack)

    def __repr__(self):
        return f'data_view({self.__data.maps.items!r})'

    def __eq__(self, other):
        return self.__data == other

    def __len__(self):
        return len(self.__data)

    def __iter__(self):
        return iter(self.__data)

    def __contains__(self, key):
        return key in self.__data

    def __getitem__(self, key):
        return self.__data[key]

    def __setitem__(self, key, value):
        self.__data[key] = value

    def __delitem__(self, key):
        del self.__data[key]

    def __getattr__(self, key):
        try:
            return self.__data[key]
        except KeyError:
            pass

        # Raise the standard exception.
        return super().__getattr__(key)

    def __setattr__(self, key, value):
        self.__data[key] = value

    def __delattr__(self, key):
        try:
            del self.__data[key]
            return
        except KeyError:
            pass

        # Raise the standard exception.
        return super().__delattr__(key)

class nested_data_view:

    def __init__(self, data_stack):
        self._data_stack = data_stack

    def __repr__(self):
        return f'nested_data_view({self._data_stack})'

    def __getitem__(self, key):
        if isinstance(key, tuple):
            hits = []

            for i, d in enumerate(self._data_stack):
                if any(k in d for k in key):
                    v = self.flatten(i)

                    try:
                        hit = tuple(v[k] for k in key)
                        hits.append(hit)
                    except KeyError:
                        pass

        else:
            hits = [
                    d[key]
                    for d in self._data_stack
                    if key in d
            ]

        if not hits:
            raise KeyError(key)

        return hits

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key) from None

    def flatten(self, layer):
        if layer < 0:
            layer += len(self._data_stack)
        return data_view(self._data_stack[:layer+1])

class info_view(iadd_mixin, MutableSequence):

    def __init__(self, info_pairs):
        self._info_pairs = info_pairs  # list of (index, template) pairs

    def __repr__(self):
        return f'info_view({self._info_pairs})'

    def __eq__(self, other):
        return list(self) == other

    def __getitem__(self, i):
        return self._info_pairs[i][1]

    def __setitem__(self, i, value):
        if isinstance(i, slice):
            self._info_pairs[i] = [(-1, v) for v in value]
        else:
            self._info_pairs[i] = -1, value

    def __delitem__(self, i):
        del self._info_pairs[i]

    def __len__(self):
        return len(self._info_pairs)

    def insert(self, i, value):
        self._info_pairs.insert(i, (-1, value))

    def layers(self):
        return self._info_pairs

class reverse_view(Sequence):

    def __init__(self, items):
        self.items = items

    def __repr__(self):
        items_str = ', '.join(repr(x) for x in self)
        return f'[{items_str}]'

    def __getitem__(self, i):
        return self.items[-i - 1]

    def __len__(self):
        return len(self.items)





#!/usr/bin/env python3

import pytest
from tidyexc import Error

STRING_TYPES = [
        dict(
            template="Fixed string",
            data={},
            expected="Fixed string",
        ),
        dict(
            template="Format string: {attr}",
            data=dict(attr="value"),
            expected="Format string: value",
        ),
        dict(
            template=lambda e: f"Callable: {e.attr}",
            data=dict(attr="value"),
            expected="Callable: value",
        ),
]

def test_raise():
    with pytest.raises(Error, match="Brief"):
        raise Error("Brief")

@pytest.mark.parametrize("s", STRING_TYPES)
def test_brief(s):
    e = Error(s['template'], **s['data'])

    assert e.brief_str == s['expected']
    assert e.info_strs == []
    assert e.blame_strs == []
    assert e.hint_strs == []

    assert str(e) == f"""\
{s['expected']}
"""

@pytest.mark.parametrize("s", STRING_TYPES)
def test_info(s):
    e = Error("Brief", **s['data'])

    e.info += s['template']
    e.info += "Second line"

    with pytest.raises(AttributeError, match=r"did you mean to use `\+\=`"):
        e.info = "Direct assignment"

    assert e.brief_str == "Brief"
    assert e.info_strs == [s['expected'], "Second line"]
    assert e.blame_strs == []
    assert e.hint_strs == []

    assert str(e) == f"""\
Brief
• {s['expected']}
• Second line
"""

@pytest.mark.parametrize("s", STRING_TYPES)
def test_blame(s):
    e = Error("Brief", **s['data'])
    e.blame += s['template']
    e.blame += "Second line"

    with pytest.raises(AttributeError, match=r"did you mean to use `\+\=`"):
        e.blame = "Direct assignment"

    assert e.brief_str == "Brief"
    assert e.info_strs == []
    assert e.blame_strs == [s['expected'], "Second line"]
    assert e.hint_strs == []

    assert str(e) == f"""\
Brief
✖ {s['expected']}
✖ Second line
"""

@pytest.mark.parametrize("s", STRING_TYPES)
def test_hints(s):
    e = Error("Brief", **s['data'])
    e.hints += s['template']
    e.hints += "Second line"

    with pytest.raises(AttributeError, match=r"did you mean to use `\+\=`"):
        e.hints = "Direct assignment"

    assert e.brief_str == "Brief"
    assert e.info_strs == []
    assert e.hint_strs == [s['expected'], "Second line"]
    assert e.blame_strs == []

    assert str(e) == f"""\
Brief
• {s['expected']}
• Second line
"""

def test_data():
    e = Error(a=1)

    assert 'a' in e.data
    assert 'b' not in e.data

    assert len(e.data) == 1

    assert e.data == {'a': 1}

    # getattr
    assert e.data.a == 1
    assert e.data['a'] == 1

    with pytest.raises(AttributeError):
        e.data.b
    with pytest.raises(KeyError):
        e.data['b']

    # setattr
    e.data.b = 2
    assert e.data == {'a': 1, 'b': 2}

    e.data['c'] = 3
    assert e.data == {'a': 1, 'b': 2, 'c': 3}

    # delattr
    del e.data.b
    assert e.data == {'a': 1, 'c': 3}

    del e.data['c']
    assert e.data == {'a': 1}

    with pytest.raises(AttributeError):
        del e.data.b
    with pytest.raises(KeyError):
        del e.data['b']

def test_data_setattr():

    class A(Error):
        pass

    with A.add_info(a=1):
        a = A(b=2)

    assert a.data == {'a': 1, 'b': 2}

    a.data = {'c': 3}
    assert a.data == {'a': 1, 'c': 3}

def test_nested_data_1():

    class A(Error):
        pass

    with A.add_info("", a=1, b=2):
        with A.add_info("", b=3, c=4):
            a = A(c=5, d=6)

    assert a.data == {'a': 1, 'b': 3, 'c': 5, 'd': 6}

    assert a.nested_data['a'] == [1]
    assert a.nested_data['b'] == [2, 3]
    assert a.nested_data['c'] == [4, 5]
    assert a.nested_data['d'] == [6]

    assert a.nested_data['a','b'] == [(1, 2), (1, 3)]
    assert a.nested_data['b','a'] == [(2, 1), (3, 1)]
    assert a.nested_data['b','c'] == [(3, 4), (3, 5)]
    assert a.nested_data['c','b'] == [(4, 3), (5, 3)]
    assert a.nested_data['c','d'] == [(5, 6)]
    assert a.nested_data['d','c'] == [(6, 5)]

def test_nested_data_2():
    # Simulate what might happen with a recursive function call (e.g. several 
    # levels defining the same exact parameters, with other levels defining 
    # unrelated parameters).  The previous test is more comprehensive, but this 
    # one is more representative.

    class A(Error):
        pass

    with A.add_info("", a=1, b=2):
        with A.add_info("", a=3, b=4):
            a = A(c=5)

    assert a.data == {'a': 3, 'b': 4, 'c': 5}

    assert a.nested_data['a'] == [1, 3]
    assert a.nested_data['b'] == [2, 4]
    assert a.nested_data['c'] == [5]

    assert a.nested_data['a','b'] == [(1, 2), (3, 4)]
    assert a.nested_data['b','a'] == [(2, 1), (4, 3)]
    assert a.nested_data['a','c'] == [(3, 5)]
    assert a.nested_data['c','a'] == [(5, 3)]
    assert a.nested_data['b','c'] == [(4, 5)]
    assert a.nested_data['c','b'] == [(5, 4)]

    assert a.nested_data['a','b','c'] == [(3, 4, 5)]

def test_iadd():
    e = Error()

    e.info += "a"
    e.info += lambda e: "b"
    e.info += ["c", "d"]
    e.info += ("e", "f")

    assert e.info_strs == ["a", "b", "c", "d", "e", "f"]

@pytest.mark.parametrize("s", STRING_TYPES)
def test_push_pop_info(s):

    class A(Error):
        pass

    class B(A):
        pass

    A.push_info(s['template'], **s['data'])
    B.push_info("x: {x}", x=1)
    B.push_info("y: {y}", y=2)
    a, b = A("Brief"), B("Brief")

    assert a.info_strs == [
            s['expected'],
    ]
    assert b.info_strs == [
            s['expected'],
            "x: 1",
            "y: 2",
    ]

    with pytest.raises(IndexError):
        Error.pop_info()

    a, b = A("Brief"), B("Brief")

    assert a.info_strs == [
            s['expected'],
    ]
    assert b.info_strs == [
            s['expected'],
            "x: 1",
            "y: 2",
    ]

    B.pop_info()
    a, b = A("Brief"), B("Brief")

    assert a.info_strs == [
            s['expected'],
    ]
    assert b.info_strs == [
            s['expected'],
            "x: 1",
    ]

    A.pop_info()
    a, b = A("Brief"), B("Brief")

    assert a.info_strs == [
    ]
    assert b.info_strs == [
            "x: 1",
    ]

@pytest.mark.parametrize("s", STRING_TYPES)
def test_push_clear_info(s):

    class A(Error):
        pass

    class B(A):
        pass

    A.push_info(s['template'], **s['data'])
    B.push_info("x: {x}", x=1)
    B.push_info("y: {y}", y=2)
    a, b = A("Brief"), B("Brief")

    assert a.info_strs == [
            s['expected'],
    ]
    assert b.info_strs == [
            s['expected'],
            "x: 1",
            "y: 2",
    ]

    Error.clear_info()
    a, b = A("Brief"), B("Brief")

    assert a.info_strs == [
            s['expected'],
    ]
    assert b.info_strs == [
            s['expected'],
            "x: 1",
            "y: 2",
    ]

    B.clear_info()
    a, b = A("Brief"), B("Brief")

    assert a.info_strs == [
            s['expected'],
    ]
    assert b.info_strs == [
            s['expected'],
    ]

    A.clear_info()
    a, b = A("Brief"), B("Brief")

    assert a.info_strs == [
    ]
    assert b.info_strs == [
    ]

@pytest.mark.parametrize("s", STRING_TYPES)
def test_add_info(s):

    class A(Error):
        pass

    class B(A):
        pass

    with A.add_info(s['template'], **s['data']):
        a1, b1 = A(), B()

    a2, b2 = A(), B()

    assert a1.info_strs == [
            s['expected']
    ]
    assert b1.info_strs == [
            s['expected']
    ]
    assert a2.info_strs == [
    ]
    assert b2.info_strs == [
    ]

    with B.add_info(s['template'], **s['data']):
        a1, b1 = A("Brief"), B("Brief")

    a2, b2 = A("Brief"), B("Brief")

    assert a1.info_strs == [
    ]
    assert b1.info_strs == [
            s['expected']
    ]
    assert a2.info_strs == [
    ]
    assert b2.info_strs == [
    ]

def test_add_info_multiple():

    class A(Error):
        pass

    with A.add_info(a=1, b=2):
        a1 = A()

    with A.add_info("{a}", a=3, b=4):
        a2 = A()

    with A.add_info("{a}", "{b}", a=5, b=6):
        a3 = A()

    assert a1.info_strs == []
    assert a2.info_strs == ["3"]
    assert a3.info_strs == ["5", "6"]

    assert a1.data == {'a': 1, 'b': 2}
    assert a2.data == {'a': 3, 'b': 4}
    assert a3.data == {'a': 5, 'b': 6}

def test_add_info_nested():

    class A(Error):
        pass

    with A.add_info("Outer a={a} b={b} c={c}", a=1, b=1, c=1):
        a1 = A(a=2, b=2)
        a1.info += "Local a={a} b={b} c={c}"

        with A.add_info("Inner a={a} b={b} c={c}", a=2, b=2):
            a2 = A(a=3)
            a2.info += "Local a={a} b={b} c={c}"

    assert a1.info_strs == [
            "Outer a=1 b=1 c=1",
            "Local a=2 b=2 c=1",
    ]
    assert a2.info_strs == [
            "Outer a=1 b=1 c=1",
            "Inner a=2 b=2 c=1",
            "Local a=3 b=2 c=1",
    ]

def test_add_info_nested_err():

    class A(Error):
        pass

    with A.add_info("Outer a={a} b={b}", a=1):
        a1 = A(b=2)

    with pytest.raises(KeyError):
        a1.info_strs

def test_put_info():
    e = Error(a=1, z=-1)
    e.info += "a={a}"
    e.put_info("a={a}")
    e.put_info("b={b}", b=2)
    e.put_info("b={b}", "c={c}", c=3)
    e.info += "z={z}"

    assert e.info_strs == [
            "a=1",
            "a=1",
            "b=2",
            "b=2",
            "c=3",
            "z=-1",
    ]

def test_put_info_nested():
    e = Error(a=1, b=1, c=1)
    e.put_info("a={a} b={b} c={c}", a=2, b=2)
    e.put_info("a={a} b={b} c={c}", a=3)

    assert e.info_strs == [
            "a=2 b=2 c=1",
            "a=3 b=2 c=1",
    ]

def test_str():
    e = Error("Brief")
    e.info += "Info"
    e.blame += "Blame"
    e.hints += "Hint"

    assert str(e) == """\
Brief
• Info
✖ Blame
• Hint
"""

def test_str_custom_bullet():

    class A(Error):
        info_bullet = 'i) '
        blame_bullet = 'x) '
        hint_bullet = '?) '

    a = A("Brief")
    a.info += "Info"
    a.blame += "Blame"
    a.hints += "Hint"

    assert str(a) == """\
Brief
i) Info
x) Blame
?) Hint
"""

def test_str_err():
    e = Error("{x}")
    e_str = str(e)

    assert "Error occurred while formatting Error:" in e_str
    assert "KeyError: 'x'" in e_str

def test_str_dict():
    # When I was using dotmap, I ran into a problem where dictionary data would 
    # be converted into a dotmap and rendered confusingly.  This test just 
    # makes sure that that behavior is eliminated.
    e = Error("{d}", d={'a': 1})
    assert str(e) == "{'a': 1}\n"

